"""
Client for interacting with the enterprise-subsidy service..
"""

__version__ = '2.2.1'

from .client import EnterpriseSubsidyAPIClient, EnterpriseSubsidyAPIClientV2, get_enterprise_subsidy_api_client
